program topas2hkl_stage2
  !
  !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  ! if .false. cal .hkl_weight file
  ! if .true. cal .hkl file
  logical, parameter :: make_hkl_file = .true.
  
  ! if .false. do not attempt to merge 100% correlated intensities
  logical, parameter :: make_multiplet = .false.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
  
  
  integer, parameter :: mref=5000 ! max intensities allowed to be read
  integer, parameter :: mcor=14
  integer, parameter :: mpar=100000
  double precision, dimension(mref) :: val_int ! holds topas intensities
  double precision, dimension(mref) :: err_int ! holds topas s.d. 
  
  double precision :: gof_value  ! used to hold gof keyword read from topas file
  
  ! inflate s.d. by gof_value*inflation_number
  double precision, parameter :: inflation_number = 0.9  
  
  integer, dimension(3,mref) :: ih ! Miller indices
  integer, dimension(mref) :: mult
  integer, dimension(mcor,mref) :: icor
  integer, dimension(mpar) :: iicor
  
  character (len = 10) IID,chextcode
  character (len = 1) chext1
  character (len = 30+4*mpar) line
  character (len=128) infile,outfile
  logical first_hc
  integer :: multiplet_counter  ! counter for multiplet
  integer :: i
  double precision, dimension(mref)  :: mean_err_int
  double precision, dimension(mref)  :: sum_val_int
  integer, dimension(mref)   :: multiplicity
  integer, dimension(mcor,mref) :: new_icor
  integer :: num_ref, i_mul
  

  ! Get input from user

  write(*,*) ' '
  write(*,*) ' '
  write(*,5001)
  5001 format(' Give TOPAS filename (including extension): ',$)
  read(*,'(q,a)') nin,infile
  !write(*,5002)
  !5002 format(' Give intensity parameter identifier (e.g. INT): ',$)
  !read(*,'(q,a)') nid,iid
  nid = 3
  iid(1:3) = "INT"
  write(*,*) ' '
  outfile=' '
  write(*,5011)
  5011 format(' Give HKL filename (including the .hkl extension): ',$)
  read(*,'(q,a)') nout,outfile


  !
  open(10,file=infile(:nin),status='old')
  open(11,file=outfile(:nout),status='unknown')

  ! read gof fit value
  
  5552 read(10,'(q,a)',end=5553) nl,line
       if (nl >= 3) then
         ii_int = index(line(1:3),"gof")
       else
         ii_int = 0
       end if
       if (ii_int.eq.0) goto 5552

       call ReadGOF(line,nl,ii_int,gof_value)
      
       ! In the off-change that the gof keyword is specified
       ! below the refined intensities
       close(10)
       open(10,file=infile(:nin),status='old')

  
  ! locate load hkl_m_d_th2 keyword

  ii_int = 0
  do while (ii_int == 0)
    read(10,'(q,a)',end=5554) nl,line
    ii_int = index(line(:nl), 'hkl_m_d_th2')
  end do


  ! read in topas intensities

  iref = 0 ! use to count the number of reflections
  10  read(10,'(q,a)',end=999) nl,line
	  ii_int=index(line(:nl),iid(:nid))
	  if (ii_int.eq.0) goto 10
      ! Still haven't located 1st intensity
  20  if (ii_int.eq.0) goto 30
	  iref = iref + 1
	  !write(*,*) iref
	  call read_TOPAS_hkl_line(line(:nl),nl,ih(1,iref),mult(iref),val_int(iref),err_int(iref))
	  
    ! setting an upper limit on max err_int to be 99999.
    ! the reason for this is to avoid 4th and 5th column to collide
    ! in the .hkl output file
    
    if (err_int(iref) > 99999.0) then
      ! scale down both the intensity and s.d. by the same amount
      
      val_int(iref) = val_int(iref) * 99999.0 / err_int(iref)
      err_int(iref) = 99999.0
    
    end if
      
	  
	  ! read in new line
	  read(10,'(q,a)',end=999) nl,line
	  ii_int = index(line(:nl),iid(:nid))
    goto 20


  ! locate A_matrix_normalized keyword

  30 nref = iref
  
     close(10)
     open(10,file=infile(:nin),status='old')  

     ic = 0
     do while (ic == 0)
     read(10,'(q,a)',end=5555) nl,line
       if (make_hkl_file) then
         ic = index(line(:nl), 'C_matrix_normalized')
       else
         ic = index(line(:nl), 'A_matrix_normalized')
       end if
     end do
 
     iref = 0


  ! read in normalized weight matrix elements

  110 read(10,'(q,a)',end=999) nl,line
	  ii_int = index(line(:nl),iid(:nid))  ! search for string contained in iid (most likely "INT") 
	  if (ii_int.eq.0) goto 110 ! look for first intensity again
	  
    !.. check which parameter number it is
	  icolon = index(line(:nl),':') !; write(*,*) icolon, line(icolon-4:icolon-1)
	  call ReadInteger(line,nl,icolon-4,icolon-1,INT1_par) 
	  npar = nref + INT1_par - 1
	  !write(*,*) npar
  120 if (ii_int.eq.0) goto 130
    !write(*,*) iref
	  iref = iref + 1
  !
	  icolon = index(line(:nl),':')
	  iend = icolon+4*npar+1
      !	write(*,*) line(icolon+1:iend)
      !	write(*,*) '>',iend-icolon,npar,iref,INT1_par,iend,nl
	  iv1 = icolon + 1 
      do i = 1, npar
       ! if (iref.gt.100) write(*,*) i,line(iv1:iv1+4)
	    read(line(iv1+1:iv1+4),*) iicor(i)
	    iv1 = iv1 + 4
	  end do
     !	write(*,'(15i4)') iref,(iicor(i),i=iref+INT1_par-1,min(iref+INT1_par+14,npar))
	  do jj = 1,14
	    icor(jj,iref)=0
	  end do
	  jjmax = min(14,npar-(iref+INT1_par-1))
	  if (jjmax.gt.0) then
	    do jj=1,jjmax
	      icor(jj,iref) = iicor(iref+INT1_par+jj-1)
	    end do
	  end if
  !
	  read(10,'(q,a)',end=999) nl,line
	  ii_int = index(line(:nl),iid(:nid))
	  goto 120
  130 continue    

    if (make_multiplet) then
 
      multiplet_counter = 1
      num_ref = 1
	    do iref = 1, nref
  	     
	      if (abs(icor(1,iref)) == 100) then
	        multiplet_counter = multiplet_counter + 1
	      else
	        ! Topas outputs the normalized weight matrix elements in integer percentage.
	        ! Thus an 100% element is likely to mean a rounded up 100% correlation.
	        ! Therefore:
    	    

    	    
	          mean_err_int(num_ref) = sum(err_int(iref-multiplet_counter+1:iref)) / multiplet_counter
    	    
	          sum_val_int(num_ref) = sum(val_int(iref-multiplet_counter+1:iref))
    	    
            do jj = 1, 14
              new_icor(jj,num_ref) = icor(jj,iref)
            end do     
    	    
	        do i_mul = 1, 14 - multiplet_counter + 1
	          if (num_ref-i_mul >= 1) then
	            new_icor(i_mul:14-multiplet_counter+1, num_ref-i_mul) = &
	              new_icor(multiplet_counter+i_mul-1:14, num_ref-i_mul)
            end if
	        end do
    	    
	        multiplicity(num_ref) = multiplet_counter
	        multiplet_counter = 1
	        num_ref = num_ref + 1
	      end if

	    end do
    	  
	    iref = 1
	    num_ref = num_ref - 1  ! needs to do this to get the true num_ref 
	    do i = 1, num_ref
	      do i_mul = 0, multiplicity(i)-1
	        if (make_hkl_file) then
	          write(11,1100) (ih(ii,iref+i_mul),ii=1,3),sum_val_int(i),mean_err_int(i), &
	            i,(new_icor(jj,i),jj=1,14)
	        else
	          write(11,1100) (ih(ii,iref+i_mul),ii=1,3),sum_val_int(i),1/mean_err_int(i), &
	            i,(new_icor(jj,i),jj=1,14)	        
	        end if    
	      end do 
	      iref = iref + multiplicity(i)
	    end do
	    
	  else
	  
	    do iref = 1, nref
	      if (make_hkl_file) then
	        write(11,1100) (ih(ii,iref),ii=1,3),val_int(iref),err_int(iref),iref,(icor(jj,iref),jj=1,14)
	      else
	        write(11,1100) (ih(ii,iref),ii=1,3),val_int(iref),1/err_int(iref),iref,(icor(jj,iref),jj=1,14)
	      end if
	    end do
	    
	  end if
	    
	1100 format(3i5,f12.4,f12.5,i6,x,14i5)
  999 continue
  
      ! What is this for: card 12 never opened!!!!!
      !close(12)
	  
  !
  !.. Now write a bespoke version of parameter_input.asc

  write(*,5500)
  5500 format(/,' Please enter the code letters for the Laue class symbols:',/,/, &		
  '     o - orthorhombic ',/,&
  '     b - monoclinic (unique axis b)',/,&
  '     c - monoclinic (unique axis c)',/,&
  '     a - monoclinic (unique axis a)',/,&
  '     t - tetragonal',/,&
  '     r - trigonal (hexagonal axis settings assumed)',/,&
  '     h - hexagonal',/,&
  '     x - trigonal and hexagonal together',/,&
  '     d - cubic')

  5501	read(*,'(q,a)') nextcode,chextcode
	  chext1=chextcode(1:1)

	  if ( .not.(chext1 .eq. 'o' .or. chext1 .eq. 'b' .or. chext1 .eq. 'c' .or. chext1 .eq. 'a' .or. &
	  chext1 .eq. 't' .or. chext1 .eq. 'r' .or. chext1 .eq. 'h' .or. chext1 .eq. 'x' .or. chext1 .eq. 'd' ) ) then
	  write(*,*) (' Unrecognised codeword - please retype')
	  goto 5501
	  end if
  !
  !
  !
	  open(13,file='parameter_input.asc',status='unknown')
	  write(13,5601)
  5601	format(/,'Below simply specify the input data file name and the crystal system of the data.',/)
	  write(13,5611)
  5611 format(/,'INPUT DATA FILE:')
	  write(13,5602) nout,outfile(:nout+4)
	  5602 format(q,a)
  !
	  write(13,5603) 
  5603 format(/,'CRYSTAL SYSTEM:')
	  write(13,5604) chext1
	  5604 format(a)
  !
	  write(13,5605) 
  5605 format(/,'--------------------------------------------------------------------------------------',/)
  !
  write(13,5610)
  5610 format(' The possible crystal system options which can be used above are:',/,/, &		
  '     o - orthorhombic ',/,&
  '     b - monoclinic (unique axis b)',/,&
  '     c - monoclinic (unique axis c)',/,&
  '     a - monoclinic (unique axis a)',/,&
  '     t - tetragonal',/,&
  '     r - trigonal (hexagonal axis settings assumed)',/,&
  '     h - hexagonal',/,&
  '     x - trigonal and hexagonal together',/,&
  '     d - cubic')
	  close(13)
  !
  

  ! create the advanced.asc file

	  open(13,file='advanced.asc',status='unknown')
	  write(13,6601)
  6601	format(/,'The parameters below can be used to affect the execution time of the program (with a',/, &
           'corresponding effect on the accuracy of the calculated numbers) and to modify the',/, &
           'errors of the powder data file, which filename is specified in parameter_input.asc',/,/, &
           'MODIFY INPUT DATA FILE',/, &
           '----------------------')    
      write(13, '(f8.4,a)') inflation_number*gof_value, '        (rescale errors, for explanation see +)'
      write(13, '(a)') ' '
      write(13, '(a)') ' '
      write(13, '(a)') 'ADJUST EXECUTION SPEED AND ACCURACY OF CALCULATED NUMBERS'
      write(13, '(a)') '--------------------------------------------------------'
      write(13, '(a)') '40              (correlation-cutoff value, for explanation see *)'
      write(13, '(a)') '500000          (max Monte Carlo iterations, for explanation see **)'
      write(13, '(a)') '2               (tolerance, for explanation see **)'
         
   	  write(13,6611)
  6611 format(/,/,'--------------------------------------------------------------------------------------',/,/, &
           'Short explanations:',/,/,/, &
           '+     multiply the standard deviations of the powder data file by this number',/, &
           '      (i.e. the square root of the diagonal elements of the correlation matrix)',/,/,/, &
           '*     the correlation-cutoff value is a threshold value. This number must be between',/, &
           '      0 and 100 and represents a percentage. If the correlation between two intensities',/, &
           '      is less than this number then the program treats such intentities as uncorrelated.',/,/,/, &  
           '**    When correlation is present between peaks the Monte Carlo integration technique is',/, &
           '      executed. max-Monte-Carlo-iterations is the maximum number of Monte Carlo iterations',/, &
           '      executed to bring the accuracy below the tolerance parameter. Like the correlation-',/, &
           '      cutoff parameter the tolerance represents a percentage value.') 
           
    close(13)
  
  !  
  
  ! picking up error from no 'gof' keyword in topas file
  stop
  
  5553 print *, " "
       print *, "---------ERROR----------"
       print *, "Topas keywork, gof, must be present in ", infile(:nin)
       print *, "Insert at the top of ", infile(:nin), "file"
       print *, "on a line by itself the Topas keyword, gof,"
       print *, "with no spaces in front of this keyword, "
       print *, "and run this program again."
       stop 
       
  5554 print *, " "
       print *, "---------ERROR----------"
       print *, "Topas keywork, hkl_m_d_th2, must be present in ", infile(:nin)
       stop       
       
  5555 print *, " "
       print *, "---------ERROR----------"
       if (make_hkl_file) then
         print *, "Topas keywork, C_matrix_normalized, must be present in ", infile(:nin)
       else
         print *, "Topas keywork, A_matrix_normalized, must be present in ", infile(:nin)
       end if
       stop            
  
end program topas2hkl_stage2
!
!
!
subroutine read_TOPAS_hkl_line(line,nl,ih,mult,val_int,err_int)
  integer ih(3)
  character (len=*) line  
  double precision :: val_int ! holds a single intensity
  double precision :: err_int ! holds a single s.d.
  
  !
  ! ... read h,k,l,m
  call ReadInteger(line,nl,1,i1,ih(1))
  call ReadInteger(line,nl,i1,i2,ih(2))
  call ReadInteger(line,nl,i2,i3,ih(3))
  call ReadInteger(line,nl,i3,i4,mult)
  !
  ii = index(line(:nl),'INT')
  call ReadIntensity(line,nl,ii,val_int,err_int)
  !  write(*,*) ih(1),ih(2),ih(3),mult,val_int,err_int
  !
end subroutine read_TOPAS_hkl_line
!
!
subroutine ReadInteger(line,nl,i1,i2,ival)
  character (len=*) line
  logical blank,integer_string
  !

  do i=i1,nl
    blank=line(i:i) .eq. ' ' .or. line(i:i) .eq. '	'
    if (.not.blank) then 
      ibeg=i
	  goto 10
    end if
  end do
  10 do i=ibeg,nl
	  integer_string=(ichar(line(i:i)) .ge. 48 .and. ichar(line(i:i)) .le. 57) .or. line(i:i).eq.'-'
    if (.not.integer_string) then
      i2=i
	  goto 20
    end if
  end do
  20 continue
  read(line(ibeg:i2-1),*) ival
  !
end subroutine ReadInteger
!
!
subroutine ReadIntensity(line,nl,ii,val_int,err_int)
  character (len=*) line
  double precision :: val_int ! holds a single intensity
  double precision :: err_int ! holds a single s.d.  
  
  logical blank
  !
  do i=ii,nl
  blank=line(i:i) .eq. ' ' .or. line(i:i) .eq. '	'
    if (blank) then
      i1=i
	  goto 10
    end if
  end do
  10 do i=i1,nl
  blank=line(i:i) .eq. ' ' .or. line(i:i) .eq. '	'
    if (.not.blank) then
      ibeg=i
	  goto 20
    end if
  end do
  20 do i=ibeg,nl
    if (line(i:i+1) .eq. '`_') then
      imid=i
	  goto 30
    end if
  end do
  30 read(line(ibeg:imid-1),*) val_int
  !write(*,*) val_int
  !
  ii=index(line(:nl),'_L')
  if (ii.eq.0) then
    nn=nl
  else
    nn=ii-1
  end if
  !write(*,*) imid+2,nn,ii
  read(line(imid+2:nn),*) err_int
  !write(*,*) nn,err_int
  !
end subroutine
!
!
subroutine ReadGOF(line,nl,ii,gof_value)
  character (len=*) line
  double precision :: gof_value
  
  logical blank
  !
  do i=ii,nl
  blank=line(i:i) .eq. ' ' .or. line(i:i) .eq. '	'
    if (blank) then
      i1=i
	  goto 10
    end if
  end do
  10 do i=i1,nl
  blank=line(i:i) .eq. ' ' .or. line(i:i) .eq. '	'
    if (.not.blank) then
      ibeg=i
	  goto 20
    end if
  end do
  20 read(line(ibeg:nl),*) gof_value
  !write(*,*) gof_value
  !
end subroutine